#ifndef zm_spline_h_included
#define zm_spline_h_included
/***************************************************************************
 *                          Zanoza Modeler v2.x                            *
 *      This is unpublished propietary source code of Zanoza Software.     *
 *      The copyright notice does not evidence any actual or intended      *
 *                    publication of such source code.                     *
 *                                                                         *
 *                Copyright (c) 2002-2003 Zanoza Software.                 *
 *                                                                         *
 *                        All rights reserved.                             *
 ***************************************************************************
 *--------------------------------------------------------------------------
 * @author:   Oleg M.
 *--------------------------------------------------------------------------
 * @purpose:  ZModeler 2.x Spline class
 *--------------------------------------------------------------------------
 * @histroy: 16.09.2003. Created
 * @histroy: 05.01.2004. Serialization added
 */
#include <zmodeler.h>
#include <scene/ISpline.h>
#include <helpers/status.h>
#include <helpers/serializeable.h>

namespace scene
{
  extern bool  g_bAutoDetail;
  extern long  g_nDetail;
  extern long  g_nMaxDetail;
  extern float g_fPrecision;

  //----------------------------------------------
  //
  // class CSpline : implements zmodeler's spline
  //
  //----------------------------------------------
  class CSpline : public scene::ISpline, 
                  public core::io::ISerializeable, 
                  public core::CErrorInfo
  {
  public:
    INTERFACE_DECLAREMAP(CSpline)
    CSpline();
    virtual ~CSpline();
  public:
  //-----------------------
  // ILight
  //-----------------------
    virtual ZRESULT isClosed(void);
    virtual ZRESULT setClosed(bool bClosed=true);
    /* Provides section end-control point in respect to closed splines. ZRESULT_FALSE is
     * returned once last section acieved. */
    virtual ZRESULT getSectionControls(scene::IControlPoint* pStart, scene::IControlPoint** ppEnd);
    /* Provides an adaptive detalization for supplied section (amount of mid-points) */
    virtual long getSectionAutoDetail(scene::IControlPoint* pStart, scene::IControlPoint* pEnd);
    /* Provides point for supplied section using supplied u [0..1] */
    virtual ZRESULT getSectionPoint(scene::IControlPoint* pStart, scene::IControlPoint* pEnd, float u, ZPoint3D& pt);
    /* Provides point for current section using supplied u [0..1] */
    virtual ZRESULT getCurrentSectionPoint(float u, ZPoint3D& pt);
    /* Provides point for the spline using supplied u [0..1] */
    virtual ZRESULT getSplinePoint(float u, float fLength, ZPoint3D& pt);
    /* Provides a control point which controls the curve pointed by u-parameter [0..1]. */
    virtual ZRESULT getSplineControlPoint(float u, scene::IControlPoint** ppPoint);
    /* when bControl true: provides amount of control points in spline; otherwise: provides
     * complete amount of points in the spline (using section detalization values). */
    virtual long numCurvePoints(bool bControl=false);
    virtual ZRESULT getSplineLength(float& fLen);
    virtual ZRESULT getSectionLength(scene::IControlPoint* pS, scene::IControlPoint* pE, float& fLen);


  public:
  //-----------------------
  // ISerializeable
  //-----------------------
    virtual ZRESULT saveDeclaration(core::io::IStream* pStream, services::IOpenSaveService* pSrv, LPCTSTR pszResolver=NULL);
    virtual ZRESULT saveData(core::io::IStream* pStream, DWORD dwID, services::IOpenSaveService* pSrv);
    virtual ZRESULT loadData(core::io::IStream* pStream, services::IOpenSaveService* pSrv, DWORD dwSize, DWORD dwVersion=0, bool bMerge=false);
    virtual ZRESULT getReCreationMode(core::io::eReCreationMode& mode);

  private:
    ZPoint3D    getSectionPoint(IControlPoint* p, float u);//helper
    long        getAutoDetail(ZPoint3D& ptS, ZPoint3D&ptE, ZVector& vS, ZVector& vE);
    inline ZPoint3D getSectionPoint(ZPoint3D& ptS, ZPoint3D&ptE, ZVector& vCS, ZVector& vCE, float u)
    {
      float u0 = u*u*u;
      float u1 = u*u*(1.0f-u);
      float u2 = u*(1.0f-u)*(1.0f-u);
      float u3 = (1.0f-u)*(1.0f-u)*(1.0f-u);

      return ptS*u3 + 3.0f*(ptS+vCS)*u2 + 3.0f*(ptE+vCE)*u1 + ptE*u0;
    }

    //overloads:
  public:
    //reverses the order
    virtual void reverseOrder()
    {
      ZSequence< ZPtr<tNodeKind> > seqItems;
      seqItems.length(m_nLength);
      int i = 1;
      lpNode ptr = m_pHead;
      //save in reverse order
      while (ptr)
      {
        seqItems[m_nLength-i] = ptr->m_pIntr;
        if (ptr->m_pIntr)
        {
          ZVector u0, u1;
          ptr->m_pIntr->getControlVector(scene::SCV_U0, u0);
          ptr->m_pIntr->getControlVector(scene::SCV_U1, u1);
          ptr->m_pIntr->setControlVector(scene::SCV_U0, u1);
          ptr->m_pIntr->setControlVector(scene::SCV_U1, u0);
        }
        i++;
        ptr = ptr->m_pNext;
      }
      ptr = m_pHead;
      i = 0;
      //load in direct order
      while (ptr)
      {
        ptr->m_pIntr = seqItems[i];
        i++;
        ptr = ptr->m_pNext;
      }
    }

  //-----------------------
  // data members
  //-----------------------
    bool            m_bClosed;
  };//CSpline

  //----------------------------------------------
  //
  // class CControlPoint : implements IControlPoint
  //
  //----------------------------------------------
  class CControlPoint : public scene::IControlPoint, public core::CErrorInfo
  {
  public:
    INTERFACE_DECLAREMAP(CControlPoint)
    CControlPoint();
    virtual ~CControlPoint();
  public:
  //-----------------------
  // IControlPoint
  //-----------------------
    virtual ZRESULT setType(eControlPoint type);
    virtual ZRESULT getType(eControlPoint& type);
    virtual ZRESULT setControlPosition(ZPoint3D pt);
    virtual ZRESULT getControlPosition(ZPoint3D& pt);
    virtual ZRESULT setControlVector(eControlVector type, ZVector v);
    virtual ZRESULT getControlVector(eControlVector type, ZVector& v);
    /* Sets segment detalization (which goes from this point to the next one). Setting detalization
     * to 0 will turn segment to line. */
    virtual ZRESULT setDetalization(long nDetail, bool bAuto = false);
    virtual ZRESULT getDetalization(long& nDetail, bool* pAuto = NULL);

  private:
    eControlPoint     m_type;
    ZVector           m_vControl[4];
    ZPoint3D          m_ptPosition;
    bool              m_bAutoDetail;
    long              m_nDetail;
  };//class

  //----------------------------------------------
  //
  // class CSharedControlPoint : implements IControlPoint
  //
  //----------------------------------------------
  class CSharedControlPoint : public CControlPoint
  {
  public:
    INTERFACE_OVERLOAD_INSTANCE_CREATOR(CSharedControlPoint)
    CSharedControlPoint();
    virtual ~CSharedControlPoint();
  public:
  //-----------------------
  // overloads
  //-----------------------
    virtual ZRESULT setControlPosition(ZPoint3D pt);
    virtual ZRESULT getControlPosition(ZPoint3D& pt);

    void setSharedControl(scene::IControlPoint* p) { m_pSharedControl = p; }
  private:
    ZPtr<scene::IControlPoint>  m_pSharedControl;
  };//class

};//scene

#endif //zm_spline_h_included

